//
//  MHIconView.m
//  Hootenanny
//
//  Created by Noah on 2/14/05.
//  Copyright 2005 __MyCompanyName__. All rights reserved.
//

#import "MHIconView.h"


@implementation MHIconView

- (int) columnWidth
{
    int w = 128;
    if ([dataSource respondsToSelector:@selector(widthOfIconsInView:)])
        w = [dataSource widthOfIconsInView:self];
    return w;
}

- (int) rowHeight
{
    int h = 128;
    if ([dataSource respondsToSelector:@selector(heightOfIconsInView:)])
        h = [dataSource heightOfIconsInView:self];
    return h;
}

- (int) numberOfColumns
{
    int n = [dataSource numberOfIconsInIconView:self];
    NSRect frame = [self frame];
    int cols = frame.size.width / [self columnWidth];
    if (n < cols) cols = n;
    return cols;
}

- (int) numberOfRows
{
    int n = [dataSource numberOfIconsInIconView:self];
    int cols = [self numberOfColumns];
    int rows = n / cols;
    // round up
    int remainder = n - (rows * cols);
    if (remainder) rows++;
    return rows;
}

- (id)initWithFrame:(NSRect)frame {
    self = [super initWithFrame:frame];
    if (self) {
        dataSource = delegate = nil;
		midDoubleClick = NO;
		currentIcons = nil;
    }
    return self;
}

- (void)drawRect:(NSRect)rect {
    // Drawing code here.
	[[NSColor whiteColor] set];
	[NSBezierPath fillRect:[self frame]];
        
	int i, n = [dataSource numberOfIconsInIconView:self];
	int height = 128, width = 128;
	int totalUsed = 0;
	if ([dataSource respondsToSelector:@selector(widthOfIconsInView:)])
		width = height = [dataSource widthOfIconsInView:self];
	int cols = [self frame].size.width / width;
	int extra = ([self frame].size.width - (cols * width)) / (double)cols;
	int col=0;
	[currentIcons release];
	currentIcons = [[NSMutableArray alloc] init];
	for (i=0;i<n;i++) {
		NSImage *image = [dataSource iconView:self iconAtIndex:i];
		if (image) {
			NSSize s = [image size];
			if (s.width > s.height)
				height = (float)width * s.height / s.width;
			else
				width = (float)height * s.width / s.height;
			NSRect srcRect = NSMakeRect(0,0,0,0);
			srcRect.size = [image size];
			NSRect dstRect = NSMakeRect(col * (width + extra), totalUsed, width, height);
			NSData *data = [NSData dataWithBytes:&dstRect length:sizeof(NSRect)];
			[currentIcons addObject:data];
			[image drawInRect:dstRect fromRect:srcRect operation:NSCompositeSourceOver fraction:1];
			col++;
			if (col == cols) {
				totalUsed += height;
				col = 0;
			}
		}
	}
}

- (void) resizeToScrollView:(NSNotification *)note
{
    NSScrollView *sview = [note object];
    if (![sview isKindOfClass:[NSScrollView class]] || [sview documentView] != self) return;
    NSRect f = NSMakeRect(0,0,0,0);
    f.size = [sview contentSize];
    [self setFrame:f];
    f.size.height = [self numberOfRows] * [self rowHeight];
    [self setFrame:f];
}

- (void) rightMouseDown:(NSEvent *)e
{
	NSPoint p = [e locationInWindow];
	p.x -= [self frame].origin.x;
	p.y -= [self frame].origin.y;
	if ([delegate respondsToSelector:@selector(iconView:rightMouseDownWithEvent:index:)])
		[delegate iconView:self rightMouseDownWithEvent:e index:[self indexOfIconAtPoint:p]];
}

- (void) mouseDown:(NSEvent *)e
{
	if ([e modifierFlags] & NSControlKeyMask) {
		[self rightMouseDown:e];
		return;
	}
	NSPoint p = [e locationInWindow];
	p.x -= [self frame].origin.x;
	p.y -= [self frame].origin.y;
	int index = [self indexOfIconAtPoint:p];
	if (index == -1) return;
	if (midDoubleClick) {
		if ([delegate respondsToSelector:@selector(iconView:doubleClickIconAtIndex:)])
			[delegate iconView:self doubleClickIconAtIndex:index];
	}
	midDoubleClick = YES;
	[self performSelector:@selector(doubleClickOff) withObject:nil afterDelay:0.5];
}

- (void) doubleClickOff
{
	midDoubleClick = NO;
}

- (id) delegate
{
	return delegate;
}

- (void) setDelegate:(id)d
{
	delegate = d;
}

- (id) dataSource
{
	return dataSource;
}

- (void) setDataSource:(id)d
{
	dataSource = d;
}

- (int) indexOfIconAtPoint:(NSPoint)p
{
	int i, n = [currentIcons count];
	for (i=0;i<n;i++) {
		NSRect r;
		const char *bytes = [[currentIcons objectAtIndex:i] bytes];
		r = *((NSRect *)bytes);
		if (r.origin.x < p.x && r.origin.x + r.size.width > p.x && r.origin.y < p.y & r.origin.y + r.size.height > p.y) {
			return i;
		}
	}
	return -1;
}

@end
