/* Copyright (c) 2004, Noah Sorscher
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	• 	Redistributions of source code must retain the above copyright
 * 	notice, this list of conditions and the following disclaimer.
 * 	• 	Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 * 	• 	Neither the name of the Noah Sorscher nor the names of its
 * 	contributors may be used to endorse or promote products derived from
 * 	this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
//
//  MHProxyMovie.m
//  Hootenanny
//
//  Created by Noah Sorscher on 5/1/05.
//  Copyright 2005 __MyCompanyName__. All rights reserved.
//

#import "MHProxyMovie.h"

@implementation NSString (test)
- (int) count
{
    NSLog(@"count");
    return 0;
}
@end


@implementation MHProxyMovie

- (MHProxyMovie *)init
{
    self = [super init];
    sortedKeys = myMovies = nil;
    added = nil;
    editingRow = -1;
    recursion = NO;
    return self;
}

- (void)dealloc
{
    [sortedKeys release];
    [myMovies release];
    [added release];
	[super dealloc];
}

- (void) addMovies:(NSArray *)movies
{
    if (!myMovies) myMovies = [[[NSMutableArray alloc] init] retain];
    [myMovies addObjectsFromArray:movies];
}

- (void)makeSortedKeys
{
    if (sortedKeys) [sortedKeys release];
    sortedKeys = [[NSMutableArray alloc] init];
    [sortedKeys retain];
    [sortedKeys addObjectsFromArray:[MHMovie supportedMetadata]];
    
    int i, j, m, n = [myMovies count];
    for (i=n;i<0;i++) {
        NSArray *allKeys = [[myMovies objectAtIndex:i] allKeys];
        m = [allKeys count];
        for (j=0;j<m;j++) {
            NSString *key = [allKeys objectAtIndex:j];
            if (![key isEqualToString:@"WLOC"] && ![sortedKeys containsObject:key]) {
                [sortedKeys addObject:key];
            }
        }
    }
    NSArray *blessed = [NSArray arrayWithObjects:@"Full Name", @"Copyright", @"Information", @"Performers", @"Album", @"Author", @"Writer", @"Director", @"Description", @"Comment", nil];
    for (i=[blessed count]-1; i>=0; i--) {
        [sortedKeys removeObject:[blessed objectAtIndex:i]];
        [sortedKeys insertObject:[blessed objectAtIndex:i] atIndex:0];
    }
}

// table view
- (int)numberOfRowsInTableView:(NSTableView *)aTableView
{
    if (!sortedKeys) [self makeSortedKeys];
    int count = [sortedKeys count];
    return count;
}

- (id)tableView:(NSTableView *)aTableView objectValueForTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex
{
    if (!sortedKeys) [self makeSortedKeys];
    NSString *value, *idString = [aTableColumn identifier];
    //NSArray *supported = [MHMovie supportedMetadata];
    if ([idString isEqualToString:@"Category"])
        return [sortedKeys objectAtIndex:rowIndex];
    NSString *key = [sortedKeys objectAtIndex:rowIndex];
    
    if (!added) return nil;
    
    NSDictionary *dict = [added objectForKey:key];
    value = [dict objectForKey:idString];
    int i, n = [[tLanguages selectedLanguages] count];
    if ([tLanguages toCopy]) {
        for (i=0;i<n && !value;i++) {
            NSString *lang = [tLanguages numberToString:[[[tLanguages selectedLanguages] objectAtIndex:i] intValue]];
            value = [dict objectForKey:lang];
        }
    }
    //NSLog(@"%@: %@, %@ - %@", key, dict, value, idString);
    return value;
}

- (void)tableView:(NSTableView *)aTableView setObjectValue:(id)anObject forTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex
{
    if (![myMovies count]) return;
    if (!sortedKeys) [self makeSortedKeys];
    //if (!anObject || [anObject isEqualToString:@""]) return;
    NSString *lang = [aTableColumn identifier];
    NSString *key = [sortedKeys objectAtIndex:rowIndex];
    
    int i, n = [myMovies count];
    for (i=0;i<n;i++) {
        [[myMovies objectAtIndex:i] setObject:anObject forKey:key language:lang];
    }
    if (!added) added = [[[NSMutableDictionary alloc] init] retain];
    [added setObject:anObject forKey:key];
}

- (BOOL) tableView: (NSTableView *) tableView shouldEditTableColumn: (NSTableColumn *) tableColumn row: (int) rowIndex
{
    editingRow = rowIndex;
    return ![[tableColumn identifier] isEqualToString:@"Category"];
}

- (void)tableView:(NSTableView *)aTableView willDisplayCell:(id)aCell forTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex
{
    if ([[aTableColumn identifier] isEqualToString:@"Category"]) {
        NSFontTraitMask mask = NSUnboldFontMask;
        if (rowIndex < 10)
            mask = NSBoldFontMask;
        [aCell setFont:[[NSFontManager sharedFontManager] convertFont:[aCell font] toHaveTrait:mask]];
    } else {
        /* This was to help avoid QT6 and below's 256 char display limit
        if (!sortedKeys) [self makeSortedKeys];
        NSDictionary *cat = [added objectForKey:[sortedKeys objectAtIndex:rowIndex]];
        NSString *value = [cat objectForKey:[aTableColumn identifier]];
        if ([value length] > 255) [aCell setTextColor:[NSColor blueColor]];
        else [aCell setTextColor:[NSColor blackColor]];
        */
    }
}

- (void)textDidChange:(NSNotification *)aNotification
{
    if (recursion) return;
    NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
    if (![defs boolForKey:@"autoComplete"]) return;
    if (editingRow < 0 || editingRow >= [sortedKeys count]) return;
    NSString *cat = [sortedKeys objectAtIndex:editingRow];
    NSTextView *box = [aNotification object];
    NSString *item, *item2, *string = [NSString stringWithString:[box string]];
    NSRange selection = [box selectedRange];
    int initialLength = [string length];
    if ([string isEqualToString:@""] || [string hasSuffix:@", "] || selection.location != initialLength) return;
    NSRange aRange = [string rangeOfString:@", " options:NSBackwardsSearch];
    if (aRange.length != 0)
        string = [string substringFromIndex:aRange.location+2];
    NSArray *list = [[defs objectForKey:@"learned"] objectForKey:cat];
    if (!list) return;
    int i;
    recursion = YES;
    for (i=[list count]-1;i>=0;i--) {
        item = [list objectAtIndex:i];
        item2 = [item stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
        if ([item2 hasPrefix:string])
            item = item2;
        else if (![item hasPrefix:string])
            item = nil;
        if (item) {
            item = [item substringFromIndex:[string length]];
            [box insertText:item];
            selection.length = [item length];
            [box setSelectedRange:selection];
            break;
        }
    }
    recursion = NO;
}

@end
