//
//  iiTableView.m
//  Hootenanny
//
//  Created by Noah Sorscher on 3/26/05.
//  Copyright 2005 __MyCompanyName__. All rights reserved.
//

#import "iiTableView.h"


@implementation iiTableView

-(void)awakeFromNib
{
    iconOn = NO, midDoubleClick = NO;
    normalSettings = nil;
	selectedItems = [[NSMutableIndexSet alloc] init];
    [super awakeFromNib];
}

- (BOOL)iconMode
{
    return iconOn;
}

- (void)setIconMode:(BOOL)flag
{
	if (iconOn == flag) return;
    iconOn = flag;
    if (flag) {
        [normalSettings release];
        normalSettings = nil;
        NSMutableDictionary *old = [[NSMutableDictionary alloc] init];
        [old setObject:[NSNumber numberWithFloat:[self rowHeight]] forKey:@"rowHeight"];
        [self setRowHeight:128];
        [old setObject:[[self tableColumns] copy] forKey:@"tableColumns"];
        while ([[self tableColumns] count])
            [self removeTableColumn:[[self tableColumns] objectAtIndex:0]];
        int i;
        for (i=0;i<4;i++) {
            NSTableColumn *col = [[[NSTableColumn alloc] initWithIdentifier:@"col"] autorelease];
            [col setWidth:128];
            [col setDataCell:[NSImageCell new]];
            [self addTableColumn:col];
        }
        if ([self headerView])
            [old setObject:[self headerView] forKey:@"headerView"];
        [self setHeaderView:nil];
        normalSettings = old;
    } else {
        float rheight = [[normalSettings objectForKey:@"rowHeight"] floatValue];
        if (!rheight) rheight = 17;
        [self setRowHeight:rheight];
        NSArray *tableColumns = [normalSettings objectForKey:@"tableColumns"];
        if ([tableColumns count]) {
            [(NSMutableArray *)[self tableColumns] removeAllObjects];
            [(NSMutableArray *)[self tableColumns] addObjectsFromArray:tableColumns];
        }
        if ([normalSettings objectForKey:@"headerView"]) {
            [self setHeaderView:[normalSettings objectForKey:@"headerView"]];
        }
		[self selectRowIndexes:selectedItems byExtendingSelection:NO];
    }
}

- (NSIndexSet *)selectedItems
{
	if (!iconOn) return [super selectedItems];
	return [[[NSIndexSet alloc] initWithIndexSet:selectedItems] autorelease];
}

- (int)selectedRow
{
	if (iconOn)
		return -1;
	return [super selectedRow];
}

- (NSIndexSet *)selectedRowIndexes
{
	if (iconOn)
		return nil;
	return [super selectedRowIndexes];
}

- (BOOL)isItemSelectedAtColumn:(int)col row:(int)row
{
	int numCols = [self numberOfColumns];
	int thisIndex = row * numCols + col;
	return [selectedItems containsIndex:thisIndex];
}

- (void)drawCellBackgroundForColumn:(int)col row:(int)row
{
	if (!iconOn) {
		[super drawCellBackgroundForColumn:col row:row];
		return;
	}
	if ([self isItemSelectedAtColumn:col row:row]) {
		int w = 75;
		NSColor *color = [self highlightColorForRow:row];
		if (color) [color set];
		else [[NSColor blueColor] set];
		NSRect cellFrame = [self smallFrameOfCellAtColumn:col row:row];
		cellFrame.size.width -= w;
		cellFrame.size.height -= w;
		cellFrame.origin.x += w/2;
		cellFrame.origin.y += w/2;
		NSBezierPath *path = [NSBezierPath bezierPathWithRect:cellFrame];
		[path setLineJoinStyle:NSRoundLineJoinStyle];
		[path setLineWidth:w];
		[path stroke];
	}
}

- (void)mouseDown:(NSEvent *)event
{
	NSPoint p;
    if ([event modifierFlags] & NSControlKeyMask)
        [self rightMouseDown:event];
    else if (iconOn) {
		if (midDoubleClick) {
			p = [event locationInWindow];
			if (p.x < doublePoint.x + 3 && p.x > doublePoint.x - 3 && p.y < doublePoint.y + 3 && p.y > doublePoint.y - 3) {
				id t = [self target];
				SEL s = [self doubleAction];
				[t performSelector:s withObject:nil afterDelay:0];
			}
		}
		p = [self convertPoint:[event locationInWindow] fromView:nil];
		int col = [self columnAtPoint:p];
		int row = [self rowAtPoint:p];
		if (row == -1 || col == -1) return;
		int thisIndex = row * [self numberOfColumns] + col;
		if ([[self dataSource] respondsToSelector:@selector(count)])
			if ([[self dataSource] count] <= thisIndex)
				return;
		if ([event modifierFlags] & NSCommandKeyMask) {
			if ([selectedItems containsIndex:thisIndex])
				[selectedItems removeIndex:thisIndex];
			else
				[selectedItems addIndex:thisIndex];
		} else {
			[selectedItems removeAllIndexes];
			[selectedItems addIndex:thisIndex];
		}
		[[self delegate] tableViewSelectionDidChange:self];
		[self setNeedsDisplay:YES];
		midDoubleClick = YES;
		doublePoint = [event locationInWindow];
		[self performSelector:@selector(doubleOff) withObject:nil afterDelay:GetDblTime() / 60.0];
    } else [super mouseDown:event];
}

- (void) doubleOff
{
	midDoubleClick = NO;
	doublePoint = NSMakePoint(NSNotFound, NSNotFound);
}

- (void)selectRowIndexes:(NSIndexSet *)indexes byExtendingSelection:(BOOL)extend
{
	if (iconOn) {
	
	} else {
		[super selectRowIndexes:indexes byExtendingSelection:extend];
		if ([selectedItems count])
			[selectedItems removeAllIndexes];
		[selectedItems addIndexes:[self selectedRowIndexes]];
	}
}

- (void)selectRow:(int)index byExtendingSelection:(BOOL)extend
{
	if (iconOn) {
	
	} else {
		[super selectRow:index byExtendingSelection:extend];
		if ([selectedItems count]) [selectedItems removeAllIndexes];
		[selectedItems addIndexes:[self selectedRowIndexes]];
	}
}

- (void)deselectRow:(int)index
{
	if (iconOn) {
		
	} else {
		[super deselectRow:index];
		if ([selectedItems count]) [selectedItems removeAllIndexes];
		[selectedItems addIndexes:[self selectedRowIndexes]];
	}
}

@end
