/* Copyright (c) 2004, Noah Sorscher
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	• 	Redistributions of source code must retain the above copyright
 * 	notice, this list of conditions and the following disclaimer.
 * 	• 	Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 * 	• 	Neither the name of the Noah Sorscher nor the names of its
 * 	contributors may be used to endorse or promote products derived from
 * 	this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
//
//  ratingsFormatter.m
//  Hootenanny
//
//  Created by Noah Sorscher on 5/1/05.
//  Copyright 2005 __MyCompanyName__. All rights reserved.
//

#import "ratingsFormatter.h"

@interface NSObject (eval)
- (float)evaluation;
@end

@implementation NSObject (eval)
- (float)evaluation
{
    if ([self respondsToSelector:@selector(floatValue)] && [self respondsToSelector:@selector(rangeOfString:)] && [self respondsToSelector:@selector(substringFromIndex:)] && [self respondsToSelector:@selector(substringToIndex:)]) 
    {
        NSString *string = (NSString *)self;
        NSRange r = [string rangeOfString:@"/"];
        if (!r.length) return [string floatValue];
        float numerator = [[string substringToIndex:r.location] floatValue];
        float denominator = [[string substringFromIndex:r.location + 1] floatValue];
        if (denominator == 0) return 0;
        return numerator / denominator;
    }
    return 0;
}
@end

@implementation ratingsFormatter

- (ratingsFormatter *)initWithColumn:(NSTableColumn *)c
{
    self = [super init];
    column = c;
    return self;
}

- (BOOL)getObjectValue:(id *)anObject forString:(NSString *)string errorDescription:(NSString **)error
{
    *anObject = string;
    return YES;
}

- (NSString *)stringForObjectValue:(id)anObject
{
    return [anObject description];
}

- (NSAttributedString *)attributedStringForObjectValue:(id)anObject withDefaultAttributes:(NSDictionary *)attributes
{
    //NSString *bpath = [[NSBundle mainBundle] pathForResource:@"blank" ofType:@"tif"];
    float multiplier = [anObject evaluation];
    if (multiplier > 1) {
        int power = 10;
        while (power < 1000000) {
            float aTry = multiplier / (float)power;
            if (aTry < 1) {
                multiplier = aTry;
                break;
            }
            power *= 10;
        }
    }
    int width = (float)[column width] * (float)multiplier;
    NSString *html = [NSString stringWithFormat:@"<HTML><HEAD></HEAD><BODY><IMG HEIGHT=10 WIDTH=%d></IMG></BODY></HTML>", width];
    NSData *data = [NSData dataWithBytes:[html cString] length:[html length]];
    return [[[NSAttributedString alloc] initWithHTML:data documentAttributes:nil] autorelease];
}

- (NSString *)editingStringForObjectValue:(id)anObject
{
    if ([anObject isKindOfClass:[NSString class]])
        return anObject;
    return [anObject description];
}

@end
